@@ -20,11 +20,14 @@ import java.io.IOException;
 import java.net.InetAddress;
 import java.net.SocketException;
 import java.nio.ByteBuffer;
-import java.util.Collections;
 import java.util.HashSet;
 import java.util.Set;
 import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.ExecutionException;
 import java.util.concurrent.Executor;
+import java.util.concurrent.ScheduledExecutorService;
+import java.util.concurrent.ScheduledFuture;
+import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicLong;
 
@@ -37,6 +40,7 @@ import javax.management.ObjectName;
 import javax.servlet.http.HttpUpgradeHandler;
 import javax.servlet.http.WebConnection;
 
+import org.apache.coyote.http11.upgrade.InternalHttpUpgradeHandler;
 import org.apache.juli.logging.Log;
 import org.apache.tomcat.InstanceManager;
 import org.apache.tomcat.util.ExceptionUtils;
@@ -73,28 +77,26 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
      * Endpoint that provides low-level network I/O - must be matched to the ProtocolHandler implementation
      * (ProtocolHandler using NIO, requires NIO Endpoint etc.).
      */
-    private final AbstractEndpoint<S, ?> endpoint;
+    private final AbstractEndpoint<S,?> endpoint;
 
 
     private Handler<S> handler;
 
 
-    private final Set<Processor> waitingProcessors = Collections
-            .newSetFromMap(new ConcurrentHashMap<Processor, Boolean>());
-
+    private final Set<Processor> waitingProcessors = ConcurrentHashMap.newKeySet();
 
     /**
-     * The timeout thread.
+     * Controller for the timeout scheduling.
      */
-    private AsyncTimeout asyncTimeout = null;
+    private ScheduledFuture<?> timeoutFuture = null;
+    private ScheduledFuture<?> monitorFuture;
 
-
-    public AbstractProtocol(AbstractEndpoint<S, ?> endpoint) {
+    public AbstractProtocol(AbstractEndpoint<S,?> endpoint) {
         this.endpoint = endpoint;
         ConnectionHandler<S> cHandler = new ConnectionHandler<>(this);
-        setHandler(cHandler);
         getEndpoint().setHandler(cHandler);
-        setSoLinger(Constants.DEFAULT_CONNECTION_LINGER);
+        setHandler(cHandler);
+        setConnectionLinger(Constants.DEFAULT_CONNECTION_LINGER);
         setTcpNoDelay(Constants.DEFAULT_TCP_NO_DELAY);
     }
 
@@ -172,15 +174,17 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
     }
 
 
+    private String clientCertProvider = null;
+
     /**
      * When client certificate information is presented in a form other than instances of
      * {@link java.security.cert.X509Certificate} it needs to be converted before it can be used and this property
      * controls which JSSE provider is used to perform the conversion. For example it is used with the AJP connectors,
      * the HTTP APR connector and with the {@link org.apache.catalina.valves.SSLValve}. If not specified, the default
      * provider will be used.
+     *
+     * @return The name of the JSSE provider to use
      */
-    protected String clientCertProvider = null;
-
     public String getClientCertProvider() {
         return clientCertProvider;
     }
@@ -190,6 +194,17 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
     }
 
 
+    private int maxHeaderCount = 100;
+
+    public int getMaxHeaderCount() {
+        return maxHeaderCount;
+    }
+
+    public void setMaxHeaderCount(int maxHeaderCount) {
+        this.maxHeaderCount = maxHeaderCount;
+    }
+
+
     @Override
     public boolean isAprRequired() {
         return false;
@@ -202,43 +217,9 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
     }
 
 
-    public AsyncTimeout getAsyncTimeout() {
-        return asyncTimeout;
-    }
-
-    /**
-     * Specifies whether the reason phrase will be sent in the response. By default a reason phrase will not be sent in
-     * the response.
-     *
-     * @deprecated This option will be removed in Tomcat 9. Reason phrase will not be sent.
-     */
-    @Deprecated
-    private boolean sendReasonPhrase = false;
-
-    /**
-     * Returns whether the reason phrase will be sent in the response. By default a reason phrase will not be sent in
-     * the response.
-     *
-     * @return whether the reason phrase will be sent
-     *
-     * @deprecated This option will be removed in Tomcat 9. Reason phrase will not be sent.
-     */
-    @Deprecated
-    public boolean getSendReasonPhrase() {
-        return sendReasonPhrase;
-    }
-
-    /**
-     * Specifies whether the reason phrase will be sent in the response. By default a reason phrase will not be sent in
-     * the response.
-     *
-     * @param sendReasonPhrase specifies whether the reason phrase will be sent
-     *
-     * @deprecated This option will be removed in Tomcat 9. Reason phrase will not be sent.
-     */
-    @Deprecated
-    public void setSendReasonPhrase(boolean sendReasonPhrase) {
-        this.sendReasonPhrase = sendReasonPhrase;
+    @Override
+    public String getId() {
+        return endpoint.getId();
     }
 
 
@@ -249,11 +230,23 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         return endpoint.getExecutor();
     }
 
+    @Override
     public void setExecutor(Executor executor) {
         endpoint.setExecutor(executor);
     }
 
 
+    @Override
+    public ScheduledExecutorService getUtilityExecutor() {
+        return endpoint.getUtilityExecutor();
+    }
+
+    @Override
+    public void setUtilityExecutor(ScheduledExecutorService utilityExecutor) {
+        endpoint.setUtilityExecutor(utilityExecutor);
+    }
+
+
     public int getMaxThreads() {
         return endpoint.getMaxThreads();
     }
@@ -289,6 +282,14 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
     }
 
 
+    public int getMaxQueueSize() {
+        return endpoint.getMaxQueueSize();
+    }
+
+    public void setMaxQueueSize(int maxQueueSize) {
+        endpoint.setMaxQueueSize(maxQueueSize);
+    }
+
     public int getAcceptCount() {
         return endpoint.getAcceptCount();
     }
@@ -297,16 +298,6 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         endpoint.setAcceptCount(acceptCount);
     }
 
-    @Deprecated
-    public int getBacklog() {
-        return endpoint.getBacklog();
-    }
-
-    @Deprecated
-    public void setBacklog(int backlog) {
-        endpoint.setBacklog(backlog);
-    }
-
 
     public boolean getTcpNoDelay() {
         return endpoint.getTcpNoDelay();
@@ -325,16 +316,6 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         endpoint.setConnectionLinger(connectionLinger);
     }
 
-    @Deprecated
-    public int getSoLinger() {
-        return endpoint.getSoLinger();
-    }
-
-    @Deprecated
-    public void setSoLinger(int soLinger) {
-        endpoint.setSoLinger(soLinger);
-    }
-
 
     /**
      * The time Tomcat will wait for a subsequent request before closing the connection. The default is
@@ -398,50 +379,31 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         endpoint.setConnectionTimeout(timeout);
     }
 
-    @Deprecated
-    public int getSoTimeout() {
-        return getConnectionTimeout();
-    }
-
-    @Deprecated
-    public void setSoTimeout(int timeout) {
-        setConnectionTimeout(timeout);
-    }
-
-    public int getMaxHeaderCount() {
-        return endpoint.getMaxHeaderCount();
-    }
-
-    public void setMaxHeaderCount(int maxHeaderCount) {
-        endpoint.setMaxHeaderCount(maxHeaderCount);
-    }
-
     public long getConnectionCount() {
         return endpoint.getConnectionCount();
     }
 
     /**
-     * Unused.
+     * NO-OP.
      *
-     * @param threadCount Ignored
+     * @param threadCount Unused
      *
-     * @deprecated This attribute is hard-coded to {@code 1} and is no longer configurable.
+     * @deprecated Will be removed in Tomcat 10.
      */
     @Deprecated
     public void setAcceptorThreadCount(int threadCount) {
-        endpoint.setAcceptorThreadCount(threadCount);
     }
 
     /**
-     * Unused.
+     * Always returns 1.
      *
-     * @return Always returns {@code 1}
+     * @return Always 1.
      *
-     * @deprecated This attribute is hard-coded to {@code 1} and is no longer configurable.
+     * @deprecated Will be removed in Tomcat 10.
      */
     @Deprecated
     public int getAcceptorThreadCount() {
-        return endpoint.getAcceptorThreadCount();
+        return 1;
     }
 
     public void setAcceptorThreadPriority(int threadPriority) {
@@ -477,38 +439,43 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
     private String getNameInternal() {
         StringBuilder name = new StringBuilder(getNamePrefix());
         name.append('-');
-        if (getAddress() != null) {
-            name.append(getAddress().getHostAddress());
-            name.append('-');
-        }
-        int port = getPortWithOffset();
-        if (port == 0) {
-            // Auto binding is in use. Check if port is known
-            name.append("auto-");
-            name.append(getNameIndex());
-            port = getLocalPort();
-            if (port != -1) {
+        String id = getId();
+        if (id != null) {
+            name.append(id);
+        } else {
+            if (getAddress() != null) {
+                name.append(getAddress().getHostAddress());
                 name.append('-');
+            }
+            int port = getPortWithOffset();
+            if (port == 0) {
+                // Auto binding is in use. Check if port is known
+                name.append("auto-");
+                name.append(getNameIndex());
+                port = getLocalPort();
+                if (port != -1) {
+                    name.append('-');
+                    name.append(port);
+                }
+            } else {
                 name.append(port);
             }
-        } else {
-            name.append(port);
         }
         return name.toString();
     }
 
 
     public void addWaitingProcessor(Processor processor) {
-        if (getLog().isDebugEnabled()) {
-            getLog().debug(sm.getString("abstractProtocol.waitingProcessor.add", processor));
+        if (getLog().isTraceEnabled()) {
+            getLog().trace(sm.getString("abstractProtocol.waitingProcessor.add", processor));
         }
         waitingProcessors.add(processor);
     }
 
 
     public void removeWaitingProcessor(Processor processor) {
-        if (getLog().isDebugEnabled()) {
-            getLog().debug(sm.getString("abstractProtocol.waitingProcessor.remove", processor));
+        if (getLog().isTraceEnabled()) {
+            getLog().trace(sm.getString("abstractProtocol.waitingProcessor.remove", processor));
         }
         waitingProcessors.remove(processor);
     }
@@ -524,12 +491,12 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
 
     // ----------------------------------------------- Accessors for sub-classes
 
-    protected AbstractEndpoint<S, ?> getEndpoint() {
+    protected AbstractEndpoint<S,?> getEndpoint() {
         return endpoint;
     }
 
 
-    protected Handler<S> getHandler() {
+    public Handler<S> getHandler() {
         return handler;
     }
 
@@ -704,20 +671,42 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         }
 
         endpoint.start();
-
-        // Start timeout thread
-        asyncTimeout = new AsyncTimeout();
-        Thread timeoutThread = new Thread(asyncTimeout, getNameInternal() + "-AsyncTimeout");
-        int priority = endpoint.getThreadPriority();
-        if (priority < Thread.MIN_PRIORITY || priority > Thread.MAX_PRIORITY) {
-            priority = Thread.NORM_PRIORITY;
-        }
-        timeoutThread.setPriority(priority);
-        timeoutThread.setDaemon(true);
-        timeoutThread.start();
+        monitorFuture = getUtilityExecutor().scheduleWithFixedDelay(() -> {
+            startAsyncTimeout();
+        }, 0, 60, TimeUnit.SECONDS);
     }
 
 
+    /**
+     * Note: The name of this method originated with the Servlet 3.0 asynchronous processing but evolved over time to
+     * represent a timeout that is triggered independently of the socket read/write timeouts.
+     */
+    protected void startAsyncTimeout() {
+        if (timeoutFuture == null || timeoutFuture.isDone()) {
+            if (timeoutFuture != null && timeoutFuture.isDone()) {
+                // There was an error executing the scheduled task, get it and log it
+                try {
+                    timeoutFuture.get();
+                } catch (InterruptedException | ExecutionException e) {
+                    getLog().error(sm.getString("abstractProtocolHandler.asyncTimeoutError"), e);
+                }
+            }
+            timeoutFuture = getUtilityExecutor().scheduleAtFixedRate(() -> {
+                long now = System.currentTimeMillis();
+                for (Processor processor : waitingProcessors) {
+                    processor.timeoutAsync(now);
+                }
+            }, 1, 1, TimeUnit.SECONDS);
+        }
+    }
+
+    protected void stopAsyncTimeout() {
+        if (timeoutFuture != null) {
+            timeoutFuture.cancel(false);
+            timeoutFuture = null;
+        }
+    }
+
     @Override
     public void pause() throws Exception {
         if (getLog().isInfoEnabled()) {
@@ -728,6 +717,11 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
     }
 
 
+    public boolean isPaused() {
+        return endpoint.isPaused();
+    }
+
+
     @Override
     public void resume() throws Exception {
         if (getLog().isInfoEnabled()) {
@@ -745,8 +739,14 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
             logPortOffset();
         }
 
-        if (asyncTimeout != null) {
-            asyncTimeout.stop();
+        if (monitorFuture != null) {
+            monitorFuture.cancel(true);
+            monitorFuture = null;
+        }
+        stopAsyncTimeout();
+        // Timeout any waiting processor
+        for (Processor processor : waitingProcessors) {
+            processor.timeoutAsync(-1);
         }
 
         endpoint.stop();
@@ -837,10 +837,11 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         }
 
 
+        @SuppressWarnings("deprecation")
         @Override
         public SocketState process(SocketWrapperBase<S> wrapper, SocketEvent status) {
-            if (getLog().isDebugEnabled()) {
-                getLog().debug(sm.getString("abstractConnectionHandler.process", wrapper.getSocket(), status));
+            if (getLog().isTraceEnabled()) {
+                getLog().trace(sm.getString("abstractConnectionHandler.process", wrapper.getSocket(), status));
             }
             if (wrapper == null) {
                 // Nothing to do. Socket has been closed.
@@ -854,8 +855,8 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
             // held by this variable will be associated with the SocketWrapper before this
             // method returns.
             Processor processor = (Processor) wrapper.takeCurrentProcessor();
-            if (getLog().isDebugEnabled()) {
-                getLog().debug(sm.getString("abstractConnectionHandler.connectionsGet", processor, socket));
+            if (getLog().isTraceEnabled()) {
+                getLog().trace(sm.getString("abstractConnectionHandler.connectionsGet", processor, socket));
             }
 
             // Timeouts are calculated on a dedicated thread and then
@@ -886,8 +887,8 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                         UpgradeProtocol upgradeProtocol = getProtocol().getNegotiatedProtocol(negotiatedProtocol);
                         if (upgradeProtocol != null) {
                             processor = upgradeProtocol.getProcessor(wrapper, getProtocol().getAdapter());
-                            if (getLog().isDebugEnabled()) {
-                                getLog().debug(sm.getString("abstractConnectionHandler.processorCreate", processor));
+                            if (getLog().isTraceEnabled()) {
+                                getLog().trace(sm.getString("abstractConnectionHandler.processorCreate", processor));
                             }
                         } else if (negotiatedProtocol.equals("http/1.1")) {
                             // Explicitly negotiated the default protocol.
@@ -915,18 +916,19 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                 }
                 if (processor == null) {
                     processor = recycledProcessors.pop();
-                    if (getLog().isDebugEnabled()) {
-                        getLog().debug(sm.getString("abstractConnectionHandler.processorPop", processor));
+                    if (getLog().isTraceEnabled()) {
+                        getLog().trace(sm.getString("abstractConnectionHandler.processorPop", processor));
                     }
                 }
                 if (processor == null) {
                     processor = getProtocol().createProcessor();
                     register(processor);
-                    if (getLog().isDebugEnabled()) {
-                        getLog().debug(sm.getString("abstractConnectionHandler.processorCreate", processor));
+                    if (getLog().isTraceEnabled()) {
+                        getLog().trace(sm.getString("abstractConnectionHandler.processorCreate", processor));
                     }
                 }
 
+                // Can switch to non-deprecated version in Tomcat 10.1.x
                 processor.setSslSupport(wrapper.getSslSupport(getProtocol().getClientCertProvider()));
 
                 SocketState state = SocketState.CLOSED;
@@ -962,8 +964,8 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                             release(processor);
                             // Create the upgrade processor
                             processor = getProtocol().createUpgradeProcessor(wrapper, upgradeToken);
-                            if (getLog().isDebugEnabled()) {
-                                getLog().debug(
+                            if (getLog().isTraceEnabled()) {
+                                getLog().trace(
                                         sm.getString("abstractConnectionHandler.upgradeCreate", processor, wrapper));
                             }
                             // Initialise the upgrade handler (which may trigger
@@ -982,6 +984,12 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                                     upgradeToken.getContextBind().unbind(false, oldCL);
                                 }
                             }
+                            if (httpUpgradeHandler instanceof InternalHttpUpgradeHandler) {
+                                if (((InternalHttpUpgradeHandler) httpUpgradeHandler).hasAsyncIO()) {
+                                    // The handler will initiate all further I/O
+                                    state = SocketState.ASYNC_IO;
+                                }
+                            }
                         }
                     }
                 } while (state == SocketState.UPGRADING);
@@ -1015,6 +1023,12 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                         longPoll(wrapper, processor);
                         getProtocol().addWaitingProcessor(processor);
                     }
+                } else if (state == SocketState.ASYNC_IO) {
+                    // Don't add sockets back to the poller.
+                    // The handler will initiate all further I/O
+                    if (status != SocketEvent.OPEN_WRITE) {
+                        getProtocol().addWaitingProcessor(processor);
+                    }
                 } else if (state == SocketState.SUSPENDED) {
                     // Don't add sockets back to the poller.
                     // The resumeProcessing() method will add this socket
@@ -1139,17 +1153,14 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                     // recycledProcessors since that pool is only for AJP or
                     // HTTP processors
                     recycledProcessors.push(processor);
-                    if (getLog().isDebugEnabled()) {
-                        getLog().debug("Pushed Processor [" + processor + "]");
+                    if (getLog().isTraceEnabled()) {
+                        getLog().trace("Pushed Processor [" + processor + "]");
                     }
                 }
             }
         }
 
 
-        /**
-         * Expected to be used by the Endpoint to release resources on socket close, errors etc.
-         */
         @Override
         public void release(SocketWrapperBase<S> socketWrapper) {
             Processor processor = (Processor) socketWrapper.takeCurrentProcessor();
@@ -1167,8 +1178,8 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                         ObjectName rpName = new ObjectName(
                                 getProtocol().getDomain() + ":type=RequestProcessor,worker=" + getProtocol().getName() +
                                         ",name=" + getProtocol().getProtocolName() + "Request" + count);
-                        if (getLog().isDebugEnabled()) {
-                            getLog().debug("Register [" + processor + "] as [" + rpName + "]");
+                        if (getLog().isTraceEnabled()) {
+                            getLog().trace("Register [" + processor + "] as [" + rpName + "]");
                         }
                         Registry.getRegistry(null, null).registerComponent(rp, rpName, null);
                         rp.setRpName(rpName);
@@ -1191,8 +1202,8 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
                         RequestInfo rp = r.getRequestProcessor();
                         rp.setGlobalProcessor(null);
                         ObjectName rpName = rp.getRpName();
-                        if (getLog().isDebugEnabled()) {
-                            getLog().debug("Unregister [" + rpName + "]");
+                        if (getLog().isTraceEnabled()) {
+                            getLog().trace("Unregister [" + rpName + "]");
                         }
                         Registry.getRegistry(null, null).unregisterComponent(rpName);
                         rp.setRpName(null);
@@ -1271,42 +1282,4 @@ public abstract class AbstractProtocol<S> implements ProtocolHandler, MBeanRegis
         }
     }
 
-
-    /**
-     * Async timeout thread
-     */
-    protected class AsyncTimeout implements Runnable {
-
-        private volatile boolean asyncTimeoutRunning = true;
-
-        /**
-         * The background thread that checks async requests and fires the timeout if there has been no activity.
-         */
-        @Override
-        public void run() {
-
-            // Loop until we receive a shutdown command
-            while (asyncTimeoutRunning) {
-                try {
-                    Thread.sleep(1000);
-                } catch (InterruptedException e) {
-                    // Ignore
-                }
-                long now = System.currentTimeMillis();
-                for (Processor processor : waitingProcessors) {
-                    processor.timeoutAsync(now);
-                }
-            }
-        }
-
-
-        protected void stop() {
-            asyncTimeoutRunning = false;
-
-            // Timeout any pending async request
-            for (Processor processor : waitingProcessors) {
-                processor.timeoutAsync(-1);
-            }
-        }
-    }
 }
