@@ -89,10 +89,10 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
     private final Http2UpgradeHandler handler;
     private final WindowAllocationManager allocationManager = new WindowAllocationManager(this);
     private final Request coyoteRequest;
-    private final Response coyoteResponse = new Response();
+    private final Response coyoteResponse;
     private final StreamInputBuffer inputBuffer;
     private final StreamOutputBuffer streamOutputBuffer = new StreamOutputBuffer();
-    private final Http2OutputBuffer http2OutputBuffer = new Http2OutputBuffer(coyoteResponse, streamOutputBuffer);
+    private final Http2OutputBuffer http2OutputBuffer;
     private final AtomicBoolean removedFromActiveCount = new AtomicBoolean(false);
 
     // State machine would be too much overhead
@@ -119,16 +119,24 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         super(handler.getConnectionId(), identifier);
         this.handler = handler;
         setWindowSize(handler.getRemoteSettings().getInitialWindowSize());
+
         if (coyoteRequest == null) {
             // HTTP/2 new request
-            this.coyoteRequest = new Request();
+            this.coyoteRequest = handler.getProtocol().popRequestAndResponse();
+            this.coyoteResponse = this.coyoteRequest.getResponse();
             this.inputBuffer = new StandardStreamInputBuffer();
             this.coyoteRequest.setInputBuffer(inputBuffer);
         } else {
             // HTTP/2 Push or HTTP/1.1 upgrade
+            /*
+             * Implementation note. The request passed in is always newly created so it is safe to recycle it for re-use
+             * in the Stream.recyle() method. Need to create a matching, new response.
+             */
             this.coyoteRequest = coyoteRequest;
+            this.coyoteResponse = new Response();
+            this.coyoteRequest.setResponse(coyoteResponse);
             this.inputBuffer =
-                    new SavedRequestStreamInputBuffer((SavedRequestInputFilter) coyoteRequest.getInputBuffer());
+                    new SavedRequestStreamInputBuffer((SavedRequestInputFilter) this.coyoteRequest.getInputBuffer());
             // Headers have been read by this point
             state.receivedStartOfHeaders();
             if (HTTP_UPGRADE_STREAM.equals(identifier)) {
@@ -148,6 +156,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             state.receivedEndOfStream();
         }
         this.coyoteRequest.setSendfile(handler.hasAsyncIO() && handler.getProtocol().getUseSendfile());
+        http2OutputBuffer = new Http2OutputBuffer(this.coyoteResponse, streamOutputBuffer);
         this.coyoteResponse.setOutputBuffer(http2OutputBuffer);
         this.coyoteRequest.setResponse(coyoteResponse);
         this.coyoteRequest.protocol().setString("HTTP/2.0");
@@ -210,9 +219,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         // Set the new state first since read and write both check this
         state.receivedReset();
         // Reads wait internally so need to call a method to break the wait()
-        if (inputBuffer != null) {
-            inputBuffer.receiveReset();
-        }
+        inputBuffer.receiveReset();
         cancelAllocationRequests();
     }
 
@@ -595,6 +602,22 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
     }
 
 
+    final void writeEarlyHints() throws IOException {
+        MimeHeaders headers = coyoteResponse.getMimeHeaders();
+        String originalStatus = headers.getHeader(":status");
+        headers.setValue(":status").setString("103");
+        try {
+            handler.writeHeaders(this, 0, headers, false, Constants.DEFAULT_HEADERS_FRAME_SIZE);
+        } finally {
+            if (originalStatus == null) {
+                headers.removeHeader(":status");
+            } else {
+                headers.setValue(":status").setString(originalStatus);
+            }
+        }
+    }
+
+
     @Override
     final String getConnectionId() {
         return handler.getConnectionId();
@@ -612,14 +635,8 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
 
 
     @Override
-    final ByteBuffer getInputByteBuffer() {
-        if (inputBuffer == null) {
-            // This must either be a push or an HTTP upgrade. Either way there
-            // should not be a request body so return a zero length ByteBuffer
-            // to trigger a flow control error.
-            return ZERO_LENGTH_BYTEBUFFER;
-        }
-        return inputBuffer.getInBuffer();
+    final ByteBuffer getInputByteBuffer(boolean create) {
+        return inputBuffer.getInBuffer(create);
     }
 
 
@@ -667,9 +684,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
                     Http2Error.PROTOCOL_ERROR);
         }
         state.receivedEndOfStream();
-        if (inputBuffer != null) {
-            inputBuffer.notifyEof();
-        }
+        inputBuffer.notifyEof();
     }
 
 
@@ -756,9 +771,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
                 handler.sendStreamReset(state, se);
 
                 cancelAllocationRequests();
-                if (inputBuffer != null) {
-                    inputBuffer.swallowUnread();
-                }
+                inputBuffer.swallowUnread();
             } catch (IOException ioe) {
                 ConnectionException ce =
                         new ConnectionException(sm.getString("stream.reset.fail", getConnectionId(), getIdAsString()),
@@ -784,13 +797,16 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         }
         int remaining;
         // May be null if stream was closed before any DATA frames were processed.
-        ByteBuffer inputByteBuffer = getInputByteBuffer();
+        ByteBuffer inputByteBuffer = getInputByteBuffer(false);
         if (inputByteBuffer == null) {
             remaining = 0;
         } else {
             remaining = inputByteBuffer.remaining();
         }
         handler.replaceStream(this, new RecycledStream(getConnectionId(), getIdentifier(), state, remaining));
+        coyoteRequest.recycle();
+        coyoteResponse.recycle();
+        handler.getProtocol().pushRequestAndResponse(coyoteRequest);
     }
 
 
@@ -999,9 +1015,9 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
             writeLock.lock();
             try {
                 /*
-                 * Need to ensure that there is exactly one call to flush even when there is no data to write. Too few calls
-                 * (i.e. zero) and the end of stream message is not sent for a completed asynchronous write. Too many calls
-                 * and the end of stream message is sent too soon and trailer headers are not sent.
+                 * Need to ensure that there is exactly one call to flush even when there is no data to write. Too few
+                 * calls (i.e. zero) and the end of stream message is not sent for a completed asynchronous write. Too
+                 * many calls and the end of stream message is sent too soon and trailer headers are not sent.
                  */
                 boolean dataInBuffer = buffer.position() > 0;
                 boolean flushed = false;
@@ -1167,7 +1183,19 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
 
         abstract void notifyEof();
 
-        abstract ByteBuffer getInBuffer();
+        /**
+         * Return, creating if necessary, the input buffer.
+         *
+         * @return The input buffer
+         *
+         * @deprecated Unused. Will be removed in Tomcat 11.
+         */
+        @Deprecated
+        ByteBuffer getInBuffer() {
+            return getInBuffer(true);
+        }
+
+        abstract ByteBuffer getInBuffer(boolean create);
 
         abstract void onDataAvailable() throws IOException;
 
@@ -1378,8 +1406,10 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
 
 
         @Override
-        final ByteBuffer getInBuffer() {
-            ensureBuffersExist();
+        final ByteBuffer getInBuffer(boolean create) {
+            if (create) {
+                ensureBuffersExist();
+            }
             return inBuffer;
         }
 
@@ -1506,7 +1536,7 @@ class Stream extends AbstractNonZeroStream implements HeaderEmitter {
         }
 
         @Override
-        ByteBuffer getInBuffer() {
+        ByteBuffer getInBuffer(boolean create) {
             return null;
         }
 
